<h2>Exceptions Details</h2>
{% for index, interview in interviews.items() %}
    {% for question, exceptions in interview.exceptions.items() %}
        {% include 'interview_details.html' %}
    {% endfor %}
{% endfor %}
