models_available = {
    "openai": [
        "gpt-3.5-turbo-1106",
        "gpt-4-0125-preview",
        "gpt-4-turbo-preview",
        "gpt-3.5-turbo-16k",
        "gpt-4-1106-preview",
        "gpt-4-turbo-2024-04-09",
        "gpt-3.5-turbo-16k-0613",
        "gpt-4o-2024-05-13",
        "gpt-4-turbo",
        "gpt-3.5-turbo-0613",
        "gpt-4",
        "gpt-4-0613",
        "gpt-3.5-turbo-0125",
        "gpt-3.5-turbo",
        "gpt-3.5-turbo-instruct",
        "gpt-3.5-turbo-instruct-0914",
        "gpt-3.5-turbo-0301",
        "gpt-4-vision-preview",
        "gpt-4-1106-vision-preview",
        "gpt-4o",
    ],
    "anthropic": [
        "claude-3-5-sonnet-20240620",
        "claude-3-opus-20240229",
        "claude-3-sonnet-20240229",
        "claude-3-haiku-20240307",
    ],
    "deep_infra": [
        "meta-llama/Llama-2-13b-chat-hf",
        "mistralai/Mixtral-8x22B-Instruct-v0.1",
        "Gryphe/MythoMax-L2-13b-turbo",
        "mistralai/Mistral-7B-Instruct-v0.1",
        "Austism/chronos-hermes-13b-v2",
        "meta-llama/Llama-2-70b-chat-hf",
        "mistralai/Mistral-7B-Instruct-v0.3",
        "meta-llama/Llama-2-7b-chat-hf",
        "Qwen/Qwen2-72B-Instruct",
        "HuggingFaceH4/zephyr-orpo-141b-A35b-v0.1",
        "cognitivecomputations/dolphin-2.6-mixtral-8x7b",
        "bigcode/starcoder2-15b",
        "microsoft/WizardLM-2-8x22B",
        "codellama/CodeLlama-70b-Instruct-hf",
        "Gryphe/MythoMax-L2-13b",
        "microsoft/WizardLM-2-7B",
        "01-ai/Yi-34B-Chat",
        "bigcode/starcoder2-15b-instruct-v0.1",
        "mistralai/Mixtral-8x7B-Instruct-v0.1",
        "openchat/openchat-3.6-8b",
        "meta-llama/Meta-Llama-3-8B-Instruct",
        "microsoft/Phi-3-medium-4k-instruct",
        "Phind/Phind-CodeLlama-34B-v2",
        "google/codegemma-7b-it",
        "mistralai/Mistral-7B-Instruct-v0.2",
        "deepinfra/airoboros-70b",
        "mistralai/Mixtral-8x22B-v0.1",
        "llava-hf/llava-1.5-7b-hf",
        "codellama/CodeLlama-34b-Instruct-hf",
        "google/gemma-1.1-7b-it",
        "lizpreciatior/lzlv_70b_fp16_hf",
        "databricks/dbrx-instruct",
        "nvidia/Nemotron-4-340B-Instruct",
        "Qwen/Qwen2-7B-Instruct",
        "meta-llama/Meta-Llama-3-70B-Instruct",
        "openchat/openchat_3.5",
    ],
    "google": [
        "gemini-1.0-pro",
        "gemini-1.0-pro-001",
        "gemini-1.0-pro-latest",
        "gemini-1.0-pro-vision-latest",
        "gemini-1.5-flash",
        "gemini-1.5-flash-001",
        "gemini-1.5-flash-001-tuning",
        "gemini-1.5-flash-002",
        "gemini-1.5-flash-8b",
        "gemini-1.5-flash-8b-001",
        "gemini-1.5-flash-8b-exp-0827",
        "gemini-1.5-flash-8b-exp-0924",
        "gemini-1.5-flash-8b-latest",
        "gemini-1.5-flash-exp-0827",
        "gemini-1.5-flash-latest",
        "gemini-1.5-pro",
        "gemini-1.5-pro-001",
        "gemini-1.5-pro-002",
        "gemini-1.5-pro-exp-0801",
        "gemini-1.5-pro-exp-0827",
        "gemini-1.5-pro-latest",
        "gemini-pro",
        "gemini-pro-vision",
    ],
    "bedrock": [
        "amazon.titan-tg1-large",
        "amazon.titan-text-lite-v1",
        "amazon.titan-text-express-v1",
        "anthropic.claude-instant-v1",
        "anthropic.claude-v2:1",
        "anthropic.claude-v2",
        "anthropic.claude-3-sonnet-20240229-v1:0",
        "anthropic.claude-3-haiku-20240307-v1:0",
        "anthropic.claude-3-opus-20240229-v1:0",
        "anthropic.claude-3-5-sonnet-20240620-v1:0",
        "cohere.command-text-v14",
        "cohere.command-r-v1:0",
        "cohere.command-r-plus-v1:0",
        "cohere.command-light-text-v14",
        "meta.llama3-8b-instruct-v1:0",
        "meta.llama3-70b-instruct-v1:0",
        "meta.llama3-1-8b-instruct-v1:0",
        "meta.llama3-1-70b-instruct-v1:0",
        "meta.llama3-1-405b-instruct-v1:0",
        "mistral.mistral-7b-instruct-v0:2",
        "mistral.mixtral-8x7b-instruct-v0:1",
        "mistral.mistral-large-2402-v1:0",
        "mistral.mistral-large-2407-v1:0",
    ],
}
